#!/usr/bin/env python
"""Create a command which looks for unsigned commits by the current user, and signs them.

This is as opposed to using 'find | xargs bzr re-sign'.
"""

import errno
import os
import re
import subprocess
import sys

from bzrlib.commands import Command, register_command, builtin_command_names
import bzrlib.config
import bzrlib.errors as errors
import bzrlib.gpg
from bzrlib.option import Option
import bzrlib.testament
import bzrlib.ui


def test_suite():
    module_names = [
        'bzrlib.plugins.signing.test_fetchsigs',
        ]
    from bzrlib.tests.TestUtil import TestLoader
    loader = TestLoader()
    return loader.loadTestsFromModuleNames(module_names)


def fetch_signatures(from_branch, to_branch, pb, dry_run=False, verbose=False):
    from_repo = from_branch.repository
    to_repo = to_branch.repository

    pulled_count = 0
    to_repo.lock_write()
    try:
        from_repo.lock_read()
        try:
            to_repo.start_write_group()
            try:
                revisions = to_repo.get_ancestry(
                    to_branch.last_revision())
                for count, rev_id in enumerate(revisions):
                    if rev_id is None:
                        continue
                    pb.update('checking', count, len(revisions))

                    if to_repo.has_signature_for_revision_id(rev_id):
                        # Already have a signature
                        continue

                    try:
                        txt = from_repo.get_signature_text(rev_id)
                    except errors.NoSuchRevision:
                        # No signature to pull
                        continue

                    if verbose:
                        pb.note(rev_id)

                    pulled_count += 1
                    if not dry_run:
                        to_repo.add_signature_text(rev_id, txt)
            except:
                to_repo.abort_write_group()
                raise
            else:
                to_repo.commit_write_group()
        finally:
            from_repo.unlock()
    finally:
        to_repo.unlock()
    pb.clear()

    return pulled_count


class cmd_pull_sigs(Command):
    """Pull missing signatures from another branch.
    """

    takes_options = [Option('dry-run',
                        help='Show what signatures would be pulled')
                    , Option('verbose',
                        help='Print the revision ids being pulled')
                    ]
    takes_args = ['from_branch?']

    def run(self, from_branch=None, dry_run=False, verbose=False):
        from bzrlib.branch import Branch

        local_branch = Branch.open_containing('.')[0]
        if from_branch is None:
            from_branch = local_branch.get_parent()
        if from_branch is None:
            raise errors.BzrCommandError('No parent, and no branch supplied')
        from_branch = Branch.open(from_branch)

        pb = bzrlib.ui.ui_factory.nested_progress_bar()

        pulled_count = fetch_signatures(from_branch, local_branch, pb,
                                        dry_run=dry_run, verbose=verbose)

        print 'Pulled %d signatures' % (pulled_count,)


class cmd_push_sigs(Command):
    """Push missing signatures from another branch.
    """

    takes_options = [Option('dry-run',
                        help='Show what signatures would be pushed')
                    , Option('verbose',
                        help='Print the revision ids being pushed')
                    ]
    takes_args = ['to_branch?']

    def run(self, to_branch=None, dry_run=False, verbose=False):
        from bzrlib.branch import Branch

        local_branch = Branch.open_containing('.')[0]
        if to_branch is None:
            to_branch = local_branch.get_submit_branch()
        if to_branch is None:
            raise errors.NoSubmitBranch(local_branch)
        to_branch = Branch.open(to_branch)

        pb = bzrlib.ui.ui_factory.nested_progress_bar()

        pushed_count = fetch_signatures(local_branch, to_branch, pb,
                                        dry_run=dry_run, verbose=verbose)

        print 'Pushed %d signatures' % (pushed_count,)


class cmd_verify_sigs(Command):
    """For any signed revision, verify the signature is valid.
    """

    takes_options = ['verbose']
    takes_args = ['location?']

    def run(self, location=None, batch=False, verbose=False):
        if location is None:
            from bzrlib.workingtree import WorkingTree
            # Open the containing directory
            wt = WorkingTree.open_containing('.')[0]
            b = wt.branch
            del wt
        else:
            # Passed in locations should be exact
            from bzrlib.branch import Branch
            b = Branch.open(location)
        repo = b.repository

        config = bzrlib.config.BranchConfig(b)

        # Technically we don't need the signing command,
        # just the verification command. But they are
        # hopefully the same program.
        gpg_command = config.gpg_signing_command()

        repo.lock_read()
        try:
            if batch:
                raise errors.BzrCommandError('--batch not implemented yet')
                # TODO: jam 20060506 This used to work, because each revision 
                #       signature was stored separately on disk. Now that 
                #       they are packed into one file, is there a way we
                #       can have gpg either verify the whole file at once,
                #       or somehow break it down into chunks?
                verify_batch(repo, command, verbose=verbose)
            else:
                command = [gpg_command, '--verify', '-']
                verify_in_order(repo, command, verbose=verbose)
        finally:
            repo.unlock()


def _verify_set(command, sigs):
    """Verify just a set of signatures

    This actually recursively calls gpg --verify-files to
    break down which signatures are actually invalid.

    This is designed around the idea that most of the signatures
    will be valid. So it tries to process them in large chunks
    at a time
    """

    if not sigs:
        import traceback
        print 'you must pass a list of signatures'
        traceback.print_stack()
        return []

    invalid = []
    cmd = command[:] + sigs[:]

    p = subprocess.Popen(cmd,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE)

    p.communicate()
    if p.returncode == 0:
        return []

    # There was a problem, we need to investigate
    if len(sigs) == 1:
        # With only 1 signature, that is obviously the invalid one
        return [(sigs[0], p.returncode)]

    if len(sigs) < 10:
        # Just iterate through the signatures, checking them one at a time
        for sig in sigs:
            invalid.extend(_verify_set(command, [sig]))
        return invalid

    # Break up the sig list into approximately equal sized chunks
    # And verify them in batch mode
    count = len(sigs) / 5
    while sigs:
        chunk = sigs[:count]
        sigs = sigs[count:]
        invalid.extend(_verify_set(command, chunk))

    return invalid

    
def verify_batch(repo, command, verbose=False):
    """Verify signatures in groups.

    This should be faster, but requires breaking up the
    set to determine which signatures failed. And has
    no method for handling unknown signatures.
    """
    sigs = []
    pb = bzrlib.ui.ui_factory.nested_progress_bar()

    n_revs = len(repo.revision_store)
    for count, rev_id in enumerate(repo.revision_store):
        pb.update('finding sigs', count, n_revs)
        if repo.revision_store.has_id(rev_id, suffix='sig'):
            # This is cheating to get the file path
            relpath = repo.revision_store._relpath(rev_id, suffixes=['sig'])
            path = repo.revision_store._transport.abspath(relpath)
            sigs.append(path)
    pb.clear()

    n_sigs = len(sigs)
    # Break it up into checking 50 at a time
    chunk_size = 50
    invalid = []
    count = 0
    while sigs:
        chunk = sigs[:chunk_size]
        sigs = sigs[chunk_size:]
        pb.update('checking', count, n_sigs)
        invalid.extend(_verify_set(command, chunk))
        count += len(chunk)
    pb.clear()

    invalid_sigs = []
    unknown_sigs = []
    unknown_return = []

    # TODO: jam 20060214 To make the output identical between batch
    #       and not batch at this point we could read in the revisions
    #       and figure out their committers, etc.
    for path, retcode in invalid:
        sig = os.path.basename(path)[:-4] #Remove .sig
        if retcode == 1:
            invalid_sigs.append(sig)
        elif retcode == 2:
            unknown_sigs.append(sig)
        else:
            unknown_return.append((sig, retcode))

    print 'Found %d revisions, %d signed (%.1f%%) / %d unsigned' \
        % (n_revs, n_sigs, 100.0*n_sigs / n_revs, (n_revs - n_sigs))

    print '%4d valid' % (n_sigs - len(invalid),)
    print '%4d invalid' % (len(invalid_sigs),)
    for sig in invalid_sigs:
        print '\t%s' % (sig,)

    print '%4d unable to verify' % (len(unknown_sigs),)
    for sig in unknown_sigs:
        print '\t%s' % (sig,)

    if unknown_return:
        print '** Unknown return codes:'
        for sig, retcode in unknown_return:
            print '%d for %s' % (retcode, sig)


def verify_in_order(repo, command, verbose=False):
    """Verify all of the signatures in a row."""
    committer_counts = {}

    unsigned_revs = 0
    signed_revs = 0
    valid_sigs = 0
    invalid_sigs = []
    unknown_sigs = []
    invalid_testament = []
    mismatched_testament = []
    mismatched_email = []

    # Looking for <user@host.com>
    email_re = re.compile(r'<(?P<email>[\w+.-]+@[\w+.-]+)>')

    revisions = repo.all_revision_ids()
    n_revs = len(revisions)
    pb = bzrlib.ui.ui_factory.nested_progress_bar()
    for count, rev_id in enumerate(revisions):
        pb.update('checking', count, n_revs)
        # TODO: jam 20060506 This is one of those places where
        #       we really need a multi api. We could pierce
        #       the abstraction and go straight for the
        #       signature.knit file, and just read the
        #       signature.kndx, to see what revisions have
        #       signatures. And then we could just grab
        #       the knit, and extract the texts from it.
        #       It would probably make all of this a
        #       lot faster, rather than all the round trips.
        #       But that would be cheating.
        if not repo.has_signature_for_revision_id(rev_id):
            unsigned_revs += 1
            continue

        signed_revs += 1

        rev = repo.get_revision(rev_id)
        sig = repo.get_signature_text(rev_id)

        m = email_re.search(rev.committer)
        if m:
            rev_email = '<' + m.group('email') + '>'
        else:
            rev_email = rev.committer
        if rev.committer in committer_counts:
            committer_counts[rev.committer] += 1
        else:
            committer_counts[rev.committer] = 1

        # We don't want gpg cluttering up the screen
        p = subprocess.Popen(command,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)

        out, err = p.communicate(sig)
        if p.returncode == 0:
            valid_sigs += 1
        elif p.returncode == 1:
            invalid_sigs.append((rev.committer, rev_id))
        elif p.returncode == 2:
            unknown_sigs.append((rev.committer, rev_id))

        if verbose:
            pb.clear()
            sys.stderr.write('Revision: %s\n' % rev_id)
            sys.stderr.write(err)
        
        # Make sure that the signed testament matches 
        # what we think it should, this should probably
        # be optional, since it takes even more time
        # TODO: jam 20060506 we could be more detailed, and check the rev_id
        #       and sha1 sum directly, so that we know what actually 
        #       failed to match
        t = bzrlib.testament.Testament.from_revision(repo, rev_id)
        short_text = t.as_short_text()
        
        short_text_loc = sig.find('bazaar-ng testament short form 1\n')
        if short_text_loc == -1:
            invalid_testament.append((rev.committer, rev_id, 'could not find testament header'))
        else:
            if short_text != sig[short_text_loc:short_text_loc+len(short_text)]:
                mismatched_testament((rev.committer, rev_id))

        # Check to make sure that the email address is actually
        # present in the gpg --verify output
        if rev_email not in err:
            mismatched_email.append((rev_id, rev.committer, err))

    pb.clear()

    print 'Found %d revisions, %d signed (%.1f%%) / %d unsigned' \
        % (n_revs, signed_revs, 100.0*signed_revs / n_revs, unsigned_revs)

    print '%4d valid' % (valid_sigs,)
    print '%4d invalid' % (len(invalid_sigs),)
    if invalid_sigs:
        d = {}
        for committer, rev_id in invalid_sigs:
            if committer in d:
                d[committer] += 1
            else:
                d[committer] = 1
        for committer in sorted(d.keys()):
            print '    %d from %s' % (d[committer], committer)

    print '%4d unable to verify' % (len(unknown_sigs),)
    if unknown_sigs:
        d = {}
        for committer, rev_id in unknown_sigs:
            if committer in d:
                d[committer] += 1
            else:
                d[committer] = 1
        for committer in sorted(d.keys()):
            print '     %d from %s' % (d[committer], committer)

    print '%4d invalid testaments (signed text could not be parsed)' % (len(invalid_testament),)
    print '%4d mismatched testaments (signed text doesn\'t match generated text)' % (len(mismatched_testament),)
    print '%4d mismatched emails (committer email not in gpg signature)' % (len(mismatched_email),)
    if mismatched_email:
        for rev_id, committer, gpg_err in mismatched_email:
            print '    Revision: %s, committed by %s, gpg output:' % (rev_id, committer)
            for line in gpg_err.split('\n'):
                print '      %s' % (line,)

    print 'Signers:'
    counters = [(count, committer)
                for committer, count in committer_counts.iteritems()]
    counters.sort()
    counters.reverse()

    for count, committer in counters:
        print '\t%4d %s' % (count, committer)


register_command(cmd_pull_sigs)
register_command(cmd_push_sigs)
register_command(cmd_verify_sigs)

