from bzrlib.progress import DummyProgress
from bzrlib.tests import TestCaseWithMemoryTransport

from bzrlib.plugins.signing import fetch_signatures


class FetchSigsTests(TestCaseWithMemoryTransport):

    def test_fetch_signatures(self):
        tree = self.make_branch_and_memory_tree('source')
        source_branch = tree.branch
        source_repo = source_branch.repository
        dest_branch = self.make_branch('dest')
        dest_repo = dest_branch.repository

        tree.lock_write()
        tree.add('')
        tree.commit('revision 1')
        rev1 = source_branch.last_revision()
        tree.commit('revision 2', allow_pointless=True)
        rev2 = source_branch.last_revision()
        tree.commit('revision 3', allow_pointless=True)
        rev3 = source_branch.last_revision()
        tree.commit('revision 4', allow_pointless=True)
        rev4 = source_branch.last_revision()

        source_repo.start_write_group()
        source_repo.add_signature_text(rev2, 'signature for revision 2')
        source_repo.commit_write_group()
        tree.unlock()

        dest_branch.pull(source_branch)

        # Sign rev 4 in the dest branch.
        dest_branch.lock_write()
        dest_repo.start_write_group()
        dest_repo.add_signature_text(
            rev4, 'signature for revision 4 in dest')
        dest_repo.commit_write_group()
        dest_branch.unlock()

        # Create a new revision in the source branch, and sign
        # revisions 3-5.
        tree.lock_write()
        tree.commit('revision 5', allow_pointless=True)
        rev5 = source_branch.last_revision()
        source_repo.start_write_group()
        source_repo.add_signature_text(rev3, 'signature for revision 3')
        source_repo.add_signature_text(
            rev4, 'signature for revision 4 in source')
        source_repo.add_signature_text(rev5, 'signature for revision 5')
        source_repo.commit_write_group()
        tree.unlock()

        pb = DummyProgress()
        self.assertEqual(fetch_signatures(source_branch, dest_branch, pb), 1)

        # Fetching signatures again pulls zero signatures.
        self.assertEqual(fetch_signatures(source_branch, dest_branch, pb), 0)

        # The destination branch head is left as is.
        self.assertEqual(dest_branch.last_revision(), rev4)

        self.assertEqual(dest_repo.has_signature_for_revision_id(rev1), False)
        self.assertEqual(dest_repo.get_signature_text(rev3),
                         'signature for revision 3')
        # Existing signatures are retained.
        self.assertEqual(dest_repo.get_signature_text(rev4),
                         'signature for revision 4 in dest')

    def test_fetch_signatures_dry_run(self):
        tree = self.make_branch_and_memory_tree('source')
        source_branch = tree.branch
        source_repo = source_branch.repository
        dest_branch = self.make_branch('dest')

        tree.lock_write()
        tree.add('')
        tree.commit('revision 1')
        rev1 = source_branch.last_revision()
        tree.unlock()

        dest_branch.pull(source_branch)

        tree.lock_write()
        source_repo.start_write_group()
        source_repo.add_signature_text(rev1, 'signature for revision 1')
        source_repo.commit_write_group()
        tree.unlock()

        pb = DummyProgress()
        self.assertEqual(fetch_signatures(source_branch, dest_branch, pb,
                                          dry_run=True), 1)

        # No signature has been pulled.
        self.assertEqual(
            dest_branch.repository.has_signature_for_revision_id(rev1), False)
